/*    $ JBOSN: drv_lan7110.c v 0.1 11-12-03 21:47 $    */

/*
 * Module Name: drv_lan7110.c
 *
 * Purpose: HMS30C7110 network device driver .
 *
 */

#include <types.h>
#include <thread.h>
#include <message.h>
#include <interrupt.h>
#include <device.h>
#include <error.h>
#include <network.h>
#include <memory.h>

#include <hal_panic.h>

#include <lib/string.h>

#include <mach/hal_interrupt.h>

#include <mach/SAMA5D3X.h>
#include <mach/SAMA5D3Xbd.h>

#include <mach/gpio.h>

#include "phy.h"
#include "drv_lan.h"


#define LAN_DBG_LVL   0


//=====================================================================================
    #define MACB_DMADESC_SIZE       8
    #define RX_RING_SIZE		    512
    #define RX_BUFFER_SIZE		    128
    #define RX_BUFFER_SIZE_SHIFT    7
    #define TX_RING_SIZE		    128
    #define TX_BUFFER_SIZE          0x600       /* 1518 rounded up */
#define DMA_EMACB_SIZE          ((MACB_DMADESC_SIZE+RX_BUFFER_SIZE)*RX_RING_SIZE + (MACB_DMADESC_SIZE+TX_BUFFER_SIZE)*TX_RING_SIZE )
//=====================================================================================


#define DEF_TX_RING_PENDING    (TX_RING_SIZE - 1)

/* Make the IP header word-aligned (the ethernet header is 14 bytes) */
#define RX_OFFSET        2

#define TX_RING_GAP(pEthNetInfo)        (TX_RING_SIZE - (pEthNetInfo)->TxPending)

#define TX_BUFFS_AVAIL(pEthNetInfo)     (((pEthNetInfo)->TxTail <= (pEthNetInfo)->TxHead) ? \
                                        (pEthNetInfo)->TxTail + DEF_TX_RING_PENDING - (pEthNetInfo)->TxHead :    \
                                        (pEthNetInfo)->TxTail - (pEthNetInfo)->TxHead - 1)

#define NEXT_TX(n)                        (((n) + 1) & (TX_RING_SIZE - 1))
#define NEXT_RX(n)                        (((n) + 1) & (RX_RING_SIZE - 1))

/* minimum number of free TX descriptors before waking up TX process */
#define MACB_TX_WAKEUP_THRESH    (TX_RING_SIZE / 4)

#define MACB_RX_INT_FLAGS    (MACB_BIT(RCOMP) | MACB_BIT(RXUBR) | MACB_BIT(ISR_ROVR))

#define DMA_ADDR_OFFSET         0x20000000

NET_DEV LAN_NetDev =
{
        NULL,
        "SAMA5D3x ENET Interface0",
        DEV_TYPE_ETHER,

        (192<<0)+(168<<8)+(11<<16)+(30<<24),            // 192.168.11.30
        (255<<0)+(255<<8)+(255<<16)+(0<<24),            // 255.255.255.128
        (192<<0)+(168<<8)+(11<<16)+(1<<24),             // 192.168.11.1
};


// net configuration
#define ETH_ALEN                6

STATIC UINT32       RxPackets = 0;
STATIC UINT32       RxBytes = 0;

STATIC UINT32       TxLen = 0;

//STATIC INT32        iNetIf;

// for driver thread
#define msWait(x)               usWait((x)*2*1000)
#define LANDEV_STACK_SIZE       (4096)

STATIC UINT32                   LANDev_Stack[LANDEV_STACK_SIZE>>2];
STATIC BOOL                     thLANDeviceDriver( UINT32 Arg );
STATIC DEVICECONFIG             LAN_DeviceConfig = {"ETH", 0, -1, NULL, NULL};

PUBLIC THREADCONFIG             MACB_LANDev_Thread = 
                                {
                                    (UINT32)&LAN_NetDev, 
                                    thLANDeviceDriver, 
                                    LANDev_Stack, 
                                    LANDEV_STACK_SIZE, 
                                    THREAD_PRIORITY_DRIVER_HIGH+3
                                };

STATIC ENET_INFO        EthNetInfo;
STATIC ENET_BUFDESC     TxBufDesc[TX_BD_DEFAULT];
STATIC ENET_BUFDESC     RxBufDesc[RX_BD_DEFAULT];


// for LAN
#define min(a,b)    ((a>b)?b:a)
#define max(a,b)    ((a<b)?b:a)

STATIC AT91PS_EMAC      pEMAC = (AT91PS_EMAC)AT91C_BASE_EMAC;

EXTERN PHY_INTERFACE DM9161A_PhyInterface;
EXTERN PHY_INTERFACE KSZ8081_PhyInterface;


#define AT91C_EMAC_PHY_SOF (0x01 << 30)
#define AT91C_EMAC_PHY_CODE (0x02 << 16)
#define AT91C_EMAC_PHY_BASE_WORD (AT91C_EMAC_PHY_SOF | AT91C_EMAC_PHY_CODE)

#define AT91C_EMAC_READ_PHY (0x02 << 28)
#define AT91C_EMAC_WRITE_PHY (0x01 << 28)


/*
 * Read value stored in a PHY register.
 * Note: MDI interface is assumed to already have been enabled.
 */
STATIC VOID
ReadPhy( AT91PS_EMAC pEMAC, unsigned char phy_addr, unsigned char address, DWORD *dwValue)
{
	pEMAC->EMAC_MAN = AT91C_EMAC_PHY_BASE_WORD | AT91C_EMAC_READ_PHY | ((phy_addr & 0x1f) << 23) | (address << 18);

	/* Wait until IDLE bit in Network Status register is cleared */
	while( !(pEMAC->EMAC_NSR & AT91C_EMAC_IDLE) );
	*dwValue = (pEMAC->EMAC_MAN & AT91C_EMAC_DATA);	
}
/*
 * Write value stored in a PHY register.
 * Note: MDI interface is assumed to already have been enabled.
 */
STATIC VOID
WritePhy(AT91PS_EMAC pEMAC, unsigned char phy_addr, unsigned char address, DWORD dwValue)
{
	pEMAC->EMAC_MAN = AT91C_EMAC_PHY_BASE_WORD | AT91C_EMAC_WRITE_PHY | ((phy_addr & 0x1f) << 23) | (address << 18) | (dwValue & 0xFFFF);

	/* Wait until IDLE bit in Network Status register is cleared */
	while (!(pEMAC->EMAC_NSR & AT91C_EMAC_IDLE));	
}

STATIC VOID
AT91F_Enable_Mdi(AT91PS_EMAC pEMAC)
{
	pEMAC->EMAC_NCR |= AT91C_EMAC_MPE;	/* enable management port */
}

/*
 * Disable the MDIO bit in the MAC control register
 */
STATIC VOID
AT91F_Disable_Mdi(AT91PS_EMAC pEMAC)
{
	pEMAC->EMAC_NCR &= ~AT91C_EMAC_MPE;	/* disable management port */
}



STATIC VOID
PhyRegEnable( PENET_INFO pEthNetInfo )
{
    //UINT32  Flags;
    UINT32  Reg;
    
    Reg = pEMAC->EMAC_NCR;
    Reg |= MACB_BIT(MPE);
    pEMAC->EMAC_NCR = Reg;
    pEMAC->EMAC_IER = MACB_BIT(MFD);

    return;
}

STATIC VOID
PhyRegDisable( PENET_INFO pEthNetInfo )
{
    //UINT32  Flags;
    UINT32  Reg;

    Reg = pEMAC->EMAC_NCR;
    Reg &= ~MACB_BIT(MPE);
    pEMAC->EMAC_NCR = Reg;
    pEMAC->EMAC_IDR = MACB_BIT(MFD);

    return;
}


STATIC UINT32
PhyRegRead( PENET_INFO pEthNetInfo, UINT32 PhyAddr, UINT32 nReg )
{
    VOLATILE UINT32      Temp;

    PhyRegEnable( pEthNetInfo );
    
    pEMAC->EMAC_MAN =   MACB_BF(SOF, MACB_MAN_SOF) |
                        MACB_BF(RW, MACB_MAN_READ) |
                        MACB_BF(PHYA, PhyAddr) |
                        MACB_BF(REGA, nReg) |
                        MACB_BF(CODE, MACB_MAN_CODE);

    //WaitForCompletion();
{
    VOLATILE UINT32  Status;
    Sleep(1);
    do
    {
        Status = pEMAC->EMAC_ISR;
    }while( !(Status & MACB_BIT(MFD)) );

}

    Temp = MACB_BFEXT(DATA, pEMAC->EMAC_MAN);

    PhyRegDisable( pEthNetInfo );


    return Temp;

//    // set read command
//    MAC_REG_WRITE( pEthNetInfo, MII_CMD, 0x40000 | (pEthNetInfo->PhyAddr << 8) | nReg );
//
//    // wait until complete read
//    do
//    {
//        MAC_REG_READ(pEthNetInfo, MII_CMD, Temp );
//    }while( Temp & 0x40000);
//    
//    MAC_REG_READ( pEthNetInfo, MII_RXDATA, Temp );
//    
//    return Temp & 0xFFFF;
}



STATIC VOID
PhyRegWrite( PENET_INFO pEthNetInfo, UINT32 PhyAddr, UINT32 nReg, UINT32 Value )
{

    PhyRegEnable(pEthNetInfo);
    
    pEMAC->EMAC_MAN =   MACB_BF(SOF, MACB_MAN_SOF) |
                        MACB_BF(RW, MACB_MAN_WRITE) |
                        MACB_BF(PHYA, PhyAddr) |
                        MACB_BF(REGA, nReg) |
                        MACB_BF(CODE, MACB_MAN_CODE)|
                        MACB_BF(DATA, Value);

    //WaitForCompletion();
{
    VOLATILE UINT32  Status;
    do
    {
        Sleep(1);
        Status = pEMAC->EMAC_ISR;
    }while( !(Status & MACB_BIT(MFD)) );

}

    PhyRegDisable(pEthNetInfo);

    return;

//    VOLATILE UINT32      Temp;
//
//    MAC_REG( pEthNetInfo, MII_TXDATA ) = Value;
//    MAC_REG( pEthNetInfo, MII_CMD ) = 0x20000 | (pEthNetInfo->PhyAddr << 8) | nReg;
//
//    // wait until complete writing
//    do
//    {
//        Temp  = MAC_REG( pEthNetInfo, MII_CMD );
//    }while( Temp & 0x20000); 
//
//    return;
}

STATIC UINT32
PhyLinkModeRead( PENET_INFO pEthNetInfo )
{
    UINT32      RetVal;

    RetVal = MAC_MODE_100MBPS | MAC_MODE_FULLDUPLEX;
    
    return RetVal;
}


STATIC BOOL
PhyLinkStatusRead( PENET_INFO pEthNetInfo )
{
    UINT32  PhyChipStatus;
    
    // read chip status
    //PhyChipStatus = PhyRegRead( pEthNetInfo, PHY_STATUS );

    return TRUE;

    if( PhyChipStatus & 0x4 )
        return TRUE;            // PHY link-up
    else
        return FALSE;           // PHY link-down
}



STATIC VOID
MACB_SetMedia( PENET_INFO pEthNetInfo, UINT32 Mode )
{
    UINT32      Reg;

    Reg = pEMAC->EMAC_NCFGR;
    Reg &= ~(MACB_BIT(SPD) | MACB_BIT(FD));
    
    if( Mode & MAC_MODE_100MBPS )
        Reg |= MACB_BIT(SPD);
    
    if( Mode & MAC_MODE_FULLDUPLEX )
        Reg |= MACB_BIT(FD);

    pEMAC->EMAC_NCFGR = Reg;

    return;
}


STATIC BOOL
PhyProbe( PENET_INFO pEthNetInfo )
{
    UINT32      PhyAddr;
    UINT16      PhyId1, PhyId2;
    UINT32      PhyId;



    
    for( PhyAddr = 0; PhyAddr < 32 ; PhyAddr++)
    {
        PhyId1 = PhyRegRead( pEthNetInfo, PhyAddr, MII_PHYSID1 );
        PhyId2 = PhyRegRead( pEthNetInfo, PhyAddr, MII_PHYSID2 );

//DEBUG(LAN_DBG_LVL,"ID = (0x%x, 0x%x)\n", PhyId1, PhyId2);

#if 0
        PhyId = (PhyId1 << 16) | (PhyId2 & 0xfff0);
        switch (PhyId)
        {
            case MII_DM9161_ID:         /* Davicom 9161: PHY_ID1 = 0x181, PHY_ID2 = B881 */
            case MII_DM9161A_ID:        /* Davicom 9161A: PHY_ID1 = 0x181, PHY_ID2 = B8A0 */
            case MII_LXT971A_ID:        /* Intel LXT971A: PHY_ID1 = 0x13, PHY_ID2 = 78E0 */
            case MII_RTL8201_ID:        /* Realtek RTL8201: PHY_ID1 = 0, PHY_ID2 = 0x8201 */
            case MII_BCM5221_ID:        /* Broadcom BCM5221: PHY_ID1 = 0x40, PHY_ID2 = 0x61e0 */
            case MII_DP83847_ID:        /* National Semiconductor DP83847:  */
            case MII_DP83848_ID:        /* National Semiconductor DP83848:  */
            case MII_AC101L_ID:         /* Altima AC101L: PHY_ID1 = 0x22, PHY_ID2 = 0x5520 */
            case MII_KS8721_ID:         /* Micrel KS8721: PHY_ID1 = 0x22, PHY_ID2 = 0x1610 */
            case MII_T78Q21x3_ID:       /* Teridian 78Q21x3: PHY_ID1 = 0x0E, PHY_ID2 = 7237 */
            case MII_LAN83C185_ID:      /* SMSC LAN83C185: PHY_ID1 = 0x0007, PHY_ID2 = 0xC0A1 */
                detected = at91ether_setup(phy_id, phy_address, pdev, ether_clk);
                break;
        }
#else
        PhyId = (PhyId1 << 16) | (PhyId2 & 0xfff0);
        if( PhyId != 0xFFFF0000 && PhyId != 0xFFFFFFF0 &&
            PhyId != 0x0000FFF0 && PhyId != 0x00000000 )
            break;
#endif

    }
    if( PhyAddr == 32 )
    {
        DEBUG1(LAN_DBG_LVL," NO PHY CHIP found\n");
        return FALSE;
    }

    kprintf("PHY CHIP found : addr = %x, ID(%04x, %04x)\n", PhyAddr, PhyId1, PhyId2);

    pEthNetInfo->PhyAddr = PhyAddr;
    pEthNetInfo->PhyId = PhyId;



    PhyRegWrite( pEthNetInfo, pEthNetInfo->PhyAddr, MII_BMCR, (0x1<<15) |PhyRegRead( pEthNetInfo, pEthNetInfo->PhyAddr, MII_BMCR ));    // reset

    Sleep(1);


    return TRUE;
}


STATIC VOID
PhySetup( PENET_INFO pEthNetInfo )
{
    UINT32  Mode;
    UINT32  Ctrl;

    PhyProbe( pEthNetInfo );

    Ctrl = PhyRegRead( pEthNetInfo, pEthNetInfo->PhyAddr, MII_BMCR );
    Mode = PhyRegRead( pEthNetInfo, pEthNetInfo->PhyAddr, MII_BMSR );
DEBUG2(LAN_DBG_LVL,"PhySetup : Ctrl = %x\n", Ctrl);
DEBUG2(LAN_DBG_LVL,"PhySetup : Mode = %x\n", Mode);

//    Mode = 0 ;
    Mode = MAC_MODE_100MBPS | MAC_MODE_FULLDUPLEX;
//    Mode = MAC_MODE_100MBPS ;

    Ctrl = 0;
    if( Mode & MAC_MODE_100MBPS )
        Ctrl |= BMCR_SPEED100;
    if( Mode & MAC_MODE_FULLDUPLEX )
        Ctrl |= BMCR_FULLDPLX;

    PhyRegWrite( pEthNetInfo, pEthNetInfo->PhyAddr, MII_BMCR, Ctrl );

    Ctrl = PhyRegRead( pEthNetInfo, pEthNetInfo->PhyAddr, MII_BMCR );
    Mode = PhyRegRead( pEthNetInfo, pEthNetInfo->PhyAddr, MII_BMSR );
DEBUG2(LAN_DBG_LVL,"PhySetup : Ctrl = %x\n", Ctrl);
DEBUG2(LAN_DBG_LVL,"PhySetup : Mode = %x\n", Mode);

    MACB_SetMedia( pEthNetInfo, Mode );

    //PhyRegWrite( pEthNetInfo, MII_ADVERTISE, ADVERTISE_CSMA | ADVERTISE_ALL );
    //PhyRegWrite( pEthNetInfo, MII_BMCR, (BMCR_SPEED100 | BMCR_ANENABLE | BMCR_ANRESTART | BMCR_FULLDPLX));

    return;
}



//*----------------------------------------------------------------------------
//* \fn    PhyInit
//* \brief This function initialise the ethernet
//*----------------------------------------------------------------------------
STATIC BOOL
PhyInit( PENET_INFO pEthNetInfo )
{
	UINT32              MacConfig;
	MAC_PHY_OP			MacPhyOp;
	PHY_INFO            PhyInfo;
	PHY_PINTERFACE      pPhyInterface;
	BYTE                PhyAddr;	
	PHY_PINTERFACE      PhyList[] = 
	{		
        &KSZ8081_PhyInterface,
		&DM9161A_PhyInterface,
	};


	MacPhyOp.ReadPhy = ReadPhy;
	MacPhyOp.WritePhy = WritePhy;
	MacPhyOp.pDeviceLocation = pEMAC;

	DEBUGMSG(LAN_DBG_LVL,"->Enter in PhyInit \n\r");

	/* Read the PHY ID registers */
	AT91F_Enable_Mdi(pEMAC);
	

	pPhyInterface = FindPhy(&MacPhyOp, PhyList, sizeof(PhyList)/sizeof(PhyList[0]), &PhyAddr);
	if( pPhyInterface == NULL )
	{
		DEBUGMSG(1,"No Phy Found\r\n");
		return -1;
	}
	DEBUGMSG(LAN_DBG_LVL,"Found Phy (%s) at address %d\r\n",pPhyInterface->pName,PhyAddr);

	//pPhyInterface->PHY_Reset(&MacPhyOp,PhyAddr);
#define GET_LINK_SPEED_TIMEOUT 10000
	
#ifdef IS_MISTRAL
	PhyInfo.bFullDuplex = FALSE;
	PhyInfo.bAutoNegociation = FALSE;
	PhyInfo.Speed = SPEED_10;
	pPhyInterface->PHY_SetConfiguration(&MacPhyOp,PhyAddr,&PhyInfo);
#else
	PhyInfo.bAutoNegociation = FALSE;
	if(pPhyInterface->PHY_GetConfiguration(&MacPhyOp,PhyAddr,&PhyInfo) == FALSE)
	{
		if (PhyInfo.bAutoNegociation == TRUE) 
		{
			// Autonegociation is started but not complete yet
		}
		else
		{	 // Autonegocitaion is not selected
			DEBUGMSG(LAN_DBG_LVL,"EMAC: Starting Autonegociation\r\n");
			pPhyInterface->PHY_StartAutoNegociation(&MacPhyOp,PhyAddr);
		}
		if (pPhyInterface->PHY_WaitForAutonegociationComplete(&MacPhyOp,PhyAddr,GET_LINK_SPEED_TIMEOUT) == FALSE)
		{
			DEBUGMSG(1,"Autonegociation failed\r\n");
			return -1;
		}
	}
#endif

	if (pPhyInterface->PHY_WaitForLink(&MacPhyOp,PhyAddr,GET_LINK_SPEED_TIMEOUT) == FALSE)
	{
		DEBUGMSG(1,"No Link\r\n");
		return -1;
	}

	if(pPhyInterface->PHY_GetConfiguration(&MacPhyOp,PhyAddr,&PhyInfo) == FALSE)
	{
		DEBUGMSG(1,"Unable to get the network configuration from the PHY\r\n");
		return -1;
	
	}

	DEBUGMSG(1,"\n\r EMAC Init : %d Mbit/s %s DUPLEX (%s)\r\n", PhyInfo.Speed,PhyInfo.bFullDuplex ? L"FULL" : L"HALF", PhyInfo.bRMII ? L"RMII" : L"MII");


	/* Read the PHY ID registers */
	AT91F_Disable_Mdi(pEMAC);
	
	

	pEMAC->EMAC_NCR  = 0;

	/* Update the MAC with the new information */
	MacConfig = pEMAC->EMAC_NCFGR & ~(AT91C_EMAC_SPD | AT91C_EMAC_FD);
	
	if( PhyInfo.Speed == SPEED_100 )
	{
		MacConfig |= AT91C_EMAC_SPD;
	}
	else if (PhyInfo.Speed == SPEED_10) 
	{
		//Nothing to do here
	}
	else
	{
		DEBUGMSG(1,"Invalid network speed\r\n");
	}

	if (PhyInfo.bFullDuplex== TRUE)	
	{
		MacConfig |= AT91C_EMAC_FD;
	}

	

	pEMAC->EMAC_NCFGR = MacConfig;
	pEMAC->EMAC_RSR  = 0xFFFFFFFF;
	pEMAC->EMAC_TSR  = 0xFFFFFFFF;
	pEMAC->EMAC_NCR  |= AT91C_EMAC_CLRSTAT; // Clear the statistics
	pEMAC->EMAC_USRIO  = (PhyInfo.bRMII ? AT91C_EMAC_RMII : 0) | AT91C_EMAC_CLKEN;
	pEMAC->EMAC_NCR  |= (AT91C_EMAC_TE | AT91C_EMAC_RE);

	//{ DWORD dummy = pEMAC->EMAC_ISR;}
	//pEMAC->EMAC_IER = 0x0000080;
	//{ DWORD dummy = pEMAC->EMAC_ISR;}


 	return 0;
}


//========================================

// reset the MAC
STATIC VOID
MacReset( PENET_INFO pEthNetInfo )
{
    UINT32      Temp;

    /*
     * Disable RX and TX (XXX: Should we halt the transmission
     * more gracefully?)
     */
    pEMAC->EMAC_NCR = 0;
    //macb_writel(bp, NCR, 0);


    /* Clear the stats registers (XXX: Update stats first?) */
    pEMAC->EMAC_NCR = MACB_BIT(CLRSTAT);
    //macb_writel(bp, NCR, MACB_BIT(CLRSTAT));

    /* Clear all status flags */
    pEMAC->EMAC_TSR = ~0x0;
    pEMAC->EMAC_RSR = ~0x0;
    //macb_writel(bp, TSR, ~0UL);
    //macb_writel(bp, RSR, ~0UL);

    /* Disable all interrupts */
    pEMAC->EMAC_IDR = ~0x0;
    Temp = pEMAC->EMAC_ISR;
    //macb_writel(bp, IDR, ~0UL);
    //macb_readl(bp, ISR);


    return;
}


// initialize the MAC
STATIC VOID
MacInit( PENET_INFO pEthNetInfo, NET_PDEV pNetDev  )
{
    UINT32  Config;

    // set HW ETHERNET ADDRESS
    pEMAC->EMAC_SA1L = (pNetDev->EthAddr[0]<<0)|
                       (pNetDev->EthAddr[1]<<8)|
                       (pNetDev->EthAddr[2]<<16)|
                       (pNetDev->EthAddr[3]<<24);
    pEMAC->EMAC_SA1H = (pNetDev->EthAddr[4]<<0)|
                       (pNetDev->EthAddr[5]<<8);


    // config the interface
    Config = pEMAC->EMAC_NCFGR & MACB_BF(CLK, -1L);
    Config |= MACB_BIT(PAE);        /* Pause Enable */
    Config |= MACB_BIT(DRFCS);        /* Discard Rx FCS */

    //if( pEthNetInfo->Dev->Flags & IFF_PROMISC)
        Config |= MACB_BIT(CAF);    /* Copy All Frames */
    //if(!(pEthNetInfo->Dev->Flags & IFF_BROADCAST))
    //    Config |= MACB_BIT(NBC);    /* No BroadCast */

    pEMAC->EMAC_NCFGR = Config;

    /* Initialize TX and RX buffers */
#if 0
    pEMAC->EMAC_RBQP = ((UINT32)pEthNetInfo->pRxRing) - OFFSET_NONCACHE;  // physical address
    pEMAC->EMAC_TBQP = ((UINT32)pEthNetInfo->pTxRing) - OFFSET_NONCACHE;  // physical address
#else
    pEMAC->EMAC_RBQP = ((UINT32)pEthNetInfo->pRxRing);  // physical address
    pEMAC->EMAC_TBQP = ((UINT32)pEthNetInfo->pTxRing);  // physical address
#endif
    /* Enable TX and RX */
    pEMAC->EMAC_NCR = MACB_BIT(RE) | MACB_BIT(TE);

    /* Enable interrupts */
    pEMAC->EMAC_IER =   MACB_BIT(RCOMP) |
                        MACB_BIT(RXUBR) |
                        MACB_BIT(ISR_TUND) |
                        MACB_BIT(ISR_RLE) |
                        MACB_BIT(TXERR) |
                        MACB_BIT(TCOMP) |
                        MACB_BIT(ISR_ROVR) |
                        MACB_BIT(HRESP);

//    UINT32          i;
//    PENET_BUFDESC   pBufferToProcess;
//
//    // init flags
//    pEthNetInfo->RxStatus = FALSE;
//    pEthNetInfo->TxStatus = FALSE;
//    
//    pEthNetInfo->nTxFrame = 0;
//    pEthNetInfo->nTxDone = 0;
//    pEthNetInfo->nTxFree = 0;
//
//    // initialize buffer descriptor access pointer
//    pEthNetInfo->pNextDeviceTx = pEthNetInfo->EthDev.pTxBufDesc;
//    pEthNetInfo->pNextDriverTx = pEthNetInfo->pNextDeviceTx;
//    pEthNetInfo->pNextTxDone = pEthNetInfo->pNextDeviceTx;
//    pEthNetInfo->pNextDeviceRx = pEthNetInfo->EthDev.pRxBufDesc;
//    pEthNetInfo->pNextDriverRx = pEthNetInfo->pNextDeviceRx;
//    pEthNetInfo->pNextRxDone = pEthNetInfo->pNextDeviceRx;
//
//    // enable ethernet interrupt 
//    // except MAC_INT_CARRIER_SENSE_DETECTED, MAC_INT_MII_SCAN
//    MAC_REG(pEthNetInfo, HBP_MAC_INT_ENABLE_REG ) = MAC_INT_ENABLED_BITS;
//
//    // Attach Rx Buffer
//    for (i = 0; i < MAC_RX_BADDR_REG_FIFO_DEPTH; i++)
//    {
//        pBufferToProcess = pEthNetInfo->pNextDeviceRx;
//        if( pBufferToProcess->Status == _B_FREE_ )
//        {
//            MAC_REG( pEthNetInfo, RX_BADDR ) = (UINT32)(pBufferToProcess->Buf);
//            pEthNetInfo->pNextDeviceRx = pBufferToProcess->pNext;
//        }
//    }

    return;
}



// start the MAC
STATIC VOID
MacStart( PENET_INFO pEthNetInfo )
{
//    UINT32      Temp;
//    //PINTC       pIntc = (PINTC)INTC_BASE_PHYS;
//
//    // clear all interrupt 
//    MAC_REG( pEthNetInfo, INT_SRC ) = ~0;
//
//    InterruptEnable( pEthNetInfo->nINTVirt );
 
//    // clear interrupt controller
//    pIntc->srcpend = ;
//    pIntc->intpend = 1<<pEthNetInfo->nINT;
//
////    // enable interrupt
////#ifndef NO_MDINT
////    #if defined(CPU_HBP2100) || defined(CPU_HBP2600)
////        SYS_INT_ENABLE (IVEC_EXT6);    /* TODO : pDrvCtrl->ivec_mdint */
////    #elif defined(CPU_HBP7110)
////        SYS_INT_ENABLE (HBP_INT_LVL_GPIO);
////    #endif
////#endif
//
//    // enable ethernet interrupts
//    pIntc->mask &= ~(1<<pEthNetInfo->nINT);
//
//    // Start Tx/Rx
//    Temp  = MAC_REG( pEthNetInfo, MAC_MODE );
//    Temp |= (MAC_MODE_RX_ENABLE | MAC_MODE_TX_ENABLE);
//    MAC_REG(pEthNetInfo, MAC_MODE ) = Temp;
//
    return;
}

// stop the MAC
STATIC VOID
MacStop( PENET_INFO pEthNetInfo )
{
//    UINT32  Temp;
//
//    // interrupt disable
//    //InterruptDisable( pEthNetInfo->nINTVirt );
//    ////pIntc->mask |= (1<<pEthNetInfo->nINT);
//
//    // stop Tx/Rx
//    Temp = MAC_REG(pEthNetInfo, MAC_MODE );
//    Temp &= ~(MAC_MODE_RX_ENABLE | MAC_MODE_TX_ENABLE);
//    MAC_REG( pEthNetInfo, MAC_MODE ) = Temp;

    return;
}





STATIC
VOID LANInitMem( PENET_INFO pEthNetInfo )
{
    UINT32          i;
    UINT32          Addr;    
    PMACB_DMADESC   pRing;

#if 0
    Addr = DMA_NC_EMACB;
#else
    //Addr = DMA_EMACB;
    Addr = (UINT32)MemPageAlloc ((DMA_EMACB_SIZE + MEMPOOL_SIZE-1)>>MEMPOOL_SHIFT);
    Addr += DMA_ADDR_OFFSET;
#endif

    pEthNetInfo->pRxBuffer = Addr;       Addr += RX_RING_SIZE*RX_BUFFER_SIZE;
    pEthNetInfo->pTxBuffer = Addr;       Addr += TX_RING_SIZE*TX_BUFFER_SIZE;
    pEthNetInfo->pRxRing = Addr;         Addr += RX_RING_SIZE*MACB_DMADESC_SIZE;
    pEthNetInfo->pTxRing = Addr;         Addr += TX_RING_SIZE*MACB_DMADESC_SIZE;

DEBUG2(LAN_DBG_LVL,"LANInitMem : pEthNetInfo->pRxBuffer = 0x%x\n", pEthNetInfo->pRxBuffer);
DEBUG2(LAN_DBG_LVL,"LANInitMem : pEthNetInfo->pTxBuffer = 0x%x\n", pEthNetInfo->pTxBuffer);
DEBUG2(LAN_DBG_LVL,"LANInitMem : pEthNetInfo->pRxRing = 0x%x\n", pEthNetInfo->pRxRing);
DEBUG2(LAN_DBG_LVL,"LANInitMem : pEthNetInfo->pTxRing = 0x%x\n", pEthNetInfo->pTxRing);


    // initialize Rx releated memory 
#if 0
    Addr = pEthNetInfo->pRxBuffer  - OFFSET_NONCACHE;
#else
    Addr = pEthNetInfo->pRxBuffer;
#endif
    for (i = 0; i < RX_RING_SIZE; i++)
    {
        pRing = pEthNetInfo->pRxRing + i;

        pRing->Addr = Addr;  // physical address
        pRing->Ctrl = 0;
        Addr += RX_BUFFER_SIZE;
    }
    pRing = pEthNetInfo->pRxRing + (RX_RING_SIZE - 1);
    pRing->Addr |= MACB_BIT(RX_WRAP);
    pEthNetInfo->RxTail = 0;

    // initialize Tx releated memory 

    for (i = 0; i < TX_RING_SIZE; i++)
    {
        pRing = pEthNetInfo->pTxRing + i;

        pRing->Addr = 0;
        pRing->Ctrl = MACB_BIT(TX_USED);
    }
    pRing = pEthNetInfo->pTxRing + (TX_RING_SIZE - 1);
    pRing->Ctrl |= MACB_BIT(TX_WRAP);

    pEthNetInfo->TxHead = pEthNetInfo->TxTail = 0;

    
//    PENET_BUFDESC   pBufDesc;
//    UINT32          i;
//
//    // setup Tx buffer descriptor 
//    pBufDesc =
//    pEthNetInfo->EthDev.pTxBufDesc = TxBufDesc[pEthNetInfo->iDev];
//    for( i = 0; i < pEthNetInfo->EthDev.nTxBufDesc - 1; i++ )
//    {
//        pBufDesc->pNext = pBufDesc + 1;
//        pBufDesc->Status = _B_FREE_;
//        pBufDesc++;
//    }
//
//    pBufDesc->pNext = pEthNetInfo->EthDev.pTxBufDesc;
//    pBufDesc->Status = _B_FREE_;
//
//    // setup Rx buffer descriptor
//    pBufDesc =
//    pEthNetInfo->EthDev.pRxBufDesc = RxBufDesc[pEthNetInfo->iDev];
//    for( i = 0; i < pEthNetInfo->EthDev.nRxBufDesc; i++ )
//    {
//        pBufDesc->pNext = pBufDesc + 1;
//        pBufDesc->Status = _B_FREE_;
//        pBufDesc++;
//    }
//    pBufDesc->pNext = pEthNetInfo->EthDev.pRxBufDesc;
//    pBufDesc->Status = _B_FREE_;

    return;
}



// This is called to initialze the ethernet low level driver.  The base address of the ethernet hardware
// is passed into the routine.  The routine will return TRUE for a successful initialization.
STATIC INT32
LANInit( PENET_INFO pEthNetInfo, NET_PDEV pNetDev )
{
    UINT32  MasterClock;
    UINT32  Config;

	const struct pio_desc hw_pio[] = {
		{"ETXD0",	AT91C_PIN_PC(0), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"ETXD1",	AT91C_PIN_PC(1), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"ERXD0",	AT91C_PIN_PC(2), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"ERXD1",	AT91C_PIN_PC(3), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"ETXEN",	AT91C_PIN_PC(4), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"ECRSDV",	AT91C_PIN_PC(5), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"ERXER",	AT91C_PIN_PC(6), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"EREFCK",	AT91C_PIN_PC(7), 0, PIO_DEFAULT, PIO_PERIPH_A},	
		{"EMDC",	AT91C_PIN_PC(8), 0, PIO_DEFAULT, PIO_PERIPH_A},						 
		{"EMDIO",	AT91C_PIN_PC(9), 0, PIO_DEFAULT, PIO_PERIPH_A},
#ifdef _NEWBD_
        {"INT",	    AT91C_PIN_PD(19), 0, PIO_INPUT, PIO_PULLUP},
#else
        {"INT",	    AT91C_PIN_PE(30), 0, PIO_INPUT, PIO_PULLUP},
#endif
	};
/*	
	const struct pio_desc hw_pio[] = {
		{"ETXD2",	AT91C_PIN_PA(6),  0, PIO_DEFAULT, PIO_PERIPH_B},
		{"ETXD3",	AT91C_PIN_PA(7),  0, PIO_DEFAULT, PIO_PERIPH_B},
		{"ERXD2",	AT91C_PIN_PA(8),  0, PIO_DEFAULT, PIO_PERIPH_B},
		{"ERXD3",	AT91C_PIN_PA(9),  0, PIO_DEFAULT, PIO_PERIPH_B},

		{"ETXD0",	AT91C_PIN_PA(10), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"ETXD1",	AT91C_PIN_PA(11), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"ERXD0",	AT91C_PIN_PA(12), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"ERXD1",	AT91C_PIN_PA(13), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"ETXEN",	AT91C_PIN_PA(14), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"ERXDV",	AT91C_PIN_PA(15), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"ERXER",	AT91C_PIN_PA(16), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"ETXCK",	AT91C_PIN_PA(17), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"EMDC",	AT91C_PIN_PA(18), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"EMDIO",	AT91C_PIN_PA(19), 0, PIO_DEFAULT, PIO_PERIPH_A},
		
		{"ETXER",	AT91C_PIN_PA(27), 0, PIO_DEFAULT, PIO_PERIPH_B},
		{"ERXCK",	AT91C_PIN_PA(28), 0, PIO_DEFAULT, PIO_PERIPH_B},
		{"ECRS",	AT91C_PIN_PA(29), 0, PIO_DEFAULT, PIO_PERIPH_B},
		{"ECOL",	AT91C_PIN_PA(30), 0, PIO_DEFAULT, PIO_PERIPH_B},

//		{"MDINTR",	AT91C_PIN_PD(5),  0, PIO_DEFAULT, PIO_PERIPH_B},
	};
    
*/
	AT91PS_PMC	pPMC;
	AT91PS_RSTC pRSTC;


	pPMC = (AT91PS_PMC)AT91C_BASE_PMC;
	pRSTC= (AT91PS_RSTC)AT91C_BASE_RSTC;
		
	// Enable Peripheral clock in PMC for  EMAC		
	pPMC->PMC_PCER1 = ((unsigned int) 1 << (SAMA5D3_ID_EMAC-32));

	// The reset of the external chips is done in the beginning of the HALInit().	
	//pio_setup(hw_pio2, sizeof(hw_pio2)/sizeof(struct pio_desc));
	//pRSTC->RSTC_RCR = (0xA5 << 24 ) | AT91C_RSTC_EXTRST;

	Sleep(200);

	pio_setup(hw_pio, sizeof(hw_pio)/sizeof(struct pio_desc));



    // interrupt for external 
    //AT91C_PIO_PE31


    //memset( EthNetInfo, 0, sizeof( ENET_INFO ) );
{
    SYSINTR_REQUEST     SysIntrRequest;
    UINT32              nSysIntr = 0;

    DEBUGMSG( LAN_DBG_LVL,"%s (%d): ++\n", __FUNCTION__, __LINE__ );

    // request sysintr    
    SysIntrRequest.nHalIrq = SAMA5D3_ID_EMAC;
    SysIntrRequest.BitMap = 0;
    SysIntrRequest.Flag = SYSINTR_REQ_FLAG_STATIC;

    if( !KernelIoControl( IOCTRL_HAL_SYSINTR_REQUEST, &SysIntrRequest, sizeof( SYSINTR_REQUEST ), &nSysIntr, sizeof(UINT32) ) )
    {
        DEBUGMSG( LAN_DBG_LVL,"%s (%d): nSysIntr = %d\n", __FUNCTION__, __LINE__, nSysIntr );
        return 0;
    }
    DEBUGMSG( LAN_DBG_LVL,"%s (%d): nSysIntr = %d\n", __FUNCTION__, __LINE__, nSysIntr );
	
    pEthNetInfo->nINTVirt = nSysIntr;
}

    pEthNetInfo->nINTPhys = SAMA5D3_ID_EMAC;
    //pEthNetInfo->nINTVirt = SYSINTR_ENET0;
    pEthNetInfo->PhyAddr = 0x1F;
    pEthNetInfo->PhyType = PHY_KS8737;
    pEthNetInfo->PhyIntfMode = MAC_MODE_MII;
    //pEthNetInfo->RegBase = MAC0_BASE_ADDR;
    //
    ////pEthNetInfo->EthDev.nSCCDev = 0;
    //pEthNetInfo->EthDev.nTxBufDesc = TX_BD_DEFAULT;
    //pEthNetInfo->EthDev.nRxBufDesc = RX_BD_DEFAULT;
    //pEthNetInfo->EthDev.pTxBufDesc = NULL;
    //pEthNetInfo->EthDev.pRxBufDesc = NULL;
    //pEthNetInfo->EthDev.TxBufSize = FRAME_MAX_AL;
    //pEthNetInfo->EthDev.RxBufSize = FRAME_MAX_AL;

    pNetDev->EthAddr[0] = 0x00;
    pNetDev->EthAddr[1] = 0x40;
    pNetDev->EthAddr[2] = 0x95;
    pNetDev->EthAddr[3] = 0x36;
    pNetDev->EthAddr[4] = 0x35;
    pNetDev->EthAddr[5] = 0x00; 



    pNetDev->BrcAddr[0] = 0xFF;
    pNetDev->BrcAddr[1] = 0xFF;
    pNetDev->BrcAddr[2] = 0xFF;
    pNetDev->BrcAddr[3] = 0xFF;
    pNetDev->BrcAddr[4] = 0xFF;
    pNetDev->BrcAddr[5] = 0xFF; 

    //DEBUG2(LAN_DBG_LVL,"%s: ", dev->name);

    // enable clock
    MasterClock = BOARD_MCK;
    if( MasterClock <= 20000000 )
        Config = MACB_BF(CLK, MACB_CLK_DIV8);
    else
    if (MasterClock <= 40000000)
        Config = MACB_BF(CLK, MACB_CLK_DIV16);
    else
    if (MasterClock <= 80000000)
        Config = MACB_BF(CLK, MACB_CLK_DIV32);
    else
        Config = MACB_BF(CLK, MACB_CLK_DIV64);

    pEMAC->EMAC_NCFGR = Config;


    if( pEthNetInfo->PhyIntfMode == MAC_MODE_MII )
        pEMAC->EMAC_USRIO = MACB_BIT(CLKEN);
    else
        pEMAC->EMAC_USRIO = MACB_BIT(RMII) | MACB_BIT(CLKEN);



    // buffer description A ACN  A
    DEBUG1(LAN_DBG_LVL,"LANInitMem\n");
    LANInitMem( pEthNetInfo );

    // reset MAC
    DEBUG1(LAN_DBG_LVL,"MacReset\n");
    MacReset( pEthNetInfo );

    // init MAC
    DEBUG1(LAN_DBG_LVL,"MacInit\n");
    MacInit( pEthNetInfo, pNetDev );

    // setup PHY mode
    DEBUG1(LAN_DBG_LVL,"PhyInit\n");
    PhyInit( pEthNetInfo );

    //DEBUG1(LAN_DBG_LVL,"PhySetup\n");
    //PhySetup( pEthNetInfo );


	{ DWORD dummy = pEMAC->EMAC_ISR;}
	pEMAC->EMAC_IER = 0x0000080;
	{ DWORD dummy = pEMAC->EMAC_ISR;}


    return 0;

}   // LANInit



// This is called to initialze the ethernet low level driver.  The base address of the ethernet hardware
// is passed into the routine.  The routine will return TRUE for a successful initialization.
STATIC BOOL
LANOpen( PENET_INFO pEthNetInfo )
{
    UINT32      PhyMode;
    UINT32      MacMode;


    // LATER INITIAILIZE THE CARRIER DETECTION and SETUP THE PHY & MAC

//    // read PHY link status
//    PhyMode = PhyLinkModeRead( pEthNetInfo );
//
//    MacMode = 0x00020630 | (MAC_MODE_RX_ALL | MAC_MODE_MII | PhyMode
//#if (ENABLE_LOOPBACK & MAC_LOOPBACK)
//               | MAC_MODE_LOOPBACK | MAC_MODE_FULLDUPLEX
//#elif (ENABLE_LOOPBACK & PHY_LOOPBACK)
//               | MAC_MODE_FULLDUPLEX
//#endif
//                    );
//
//    if( pEthNetInfo->PhyIntfMode == MAC_MODE_RMII)
//    {
//        MacMode |= MAC_MODE_RMII;
//    }
//
//    // set MAC_MODE
//    MAC_REG( pEthNetInfo, MAC_MODE ) = MacMode;
//
//    // raise the interface flags - mark the device as up
//    PhyMode = PhyLinkStatusRead( pEthNetInfo );
//      if(PhyMode)
//    {
//        /* Link UP */
//        //netif_carrier_on (dev);
//    }
//      else
//    {
//        /* Link Down */
//        //netif_carrier_off (dev);
//    }

    MacStart( pEthNetInfo );

    return TRUE;
}


STATIC BOOL
LANClose( PENET_INFO pEthNetInfo )
{

    //netif_stop_queue (dev);

    // stop the device
    MacStop( pEthNetInfo );

    // disable interrupt
    InterruptDisable( pEthNetInfo->nINTVirt);

    return 0;
}





STATIC BOOL
LANSend( PENET_INFO pEthNetInfo )
{
    UINT32          Entry;
    UINT32          Ctrl;
    UINT32          Length;
    PMACB_DMADESC   pTxRing;
    PBYTE           pPacketBuf;

DEBUG1(LAN_DBG_LVL,"LANSend \n");
    // this is hard error, log it 
    if( TX_BUFFS_AVAIL(pEthNetInfo) < 1 )
    {
        //netif_stop_queue(dev);
        return FALSE;
    }

    Entry = pEthNetInfo->TxHead;
DEBUG2(LAN_DBG_LVL,"LANSend + : Entry = %d \n", Entry);

    pPacketBuf = pEthNetInfo->pTxBuffer + Entry*TX_BUFFER_SIZE;

//DEBUG2(LAN_DBG_LVL,"LANSend : pPacketBuf = %x \n", pPacketBuf);

    // get packet data from network server
    if( !NwPacketOutput( pEthNetInfo->iNetIf, pPacketBuf, &Length ) )
        return FALSE;

DEBUG2(LAN_DBG_LVL,"LANSend : Length = %d \n", Length);

    // check the length
    Length = max( FRAME_MIN - 4, Length );
    Length = min( FRAME_MAX - 4, Length );

DEBUG2(LAN_DBG_LVL,"LANSend : Length = %d \n", Length);

    Ctrl = MACB_BF(TX_FRMLEN, Length);
    Ctrl |= MACB_BIT(TX_LAST);
    if( Entry == (TX_RING_SIZE - 1) )
        Ctrl |= MACB_BIT(TX_WRAP);

    pTxRing = pEthNetInfo->pTxRing + Entry;
#if 0
    pTxRing->Addr = pPacketBuf - OFFSET_NONCACHE;   // physical address
#else
    pTxRing->Addr = pPacketBuf;   // physical address
#endif
    pTxRing->Ctrl = Ctrl;

//DEBUG2(LAN_DBG_LVL,"LANSend : pEthNetInfo->pTxRing[Entry].Addr = %x \n", pEthNetInfo->pTxRing[Entry].Addr);
//DEBUG2(LAN_DBG_LVL,"LANSend : pEthNetInfo->pTxRing[Entry].Ctrl = %x \n", pEthNetInfo->pTxRing[Entry].Ctrl);

    Entry = NEXT_TX(Entry);
    pEthNetInfo->TxHead = Entry;

DEBUG2(LAN_DBG_LVL,"LANSend - : Entry = %d \n", Entry);

    pEMAC->EMAC_NCR = pEMAC->EMAC_NCR | MACB_BIT(TSTART);
//    if( TX_BUFFS_AVAIL(pEthNetInfo) < 1 )
//        netif_stop_queue(dev);

    return TRUE;
}


STATIC VOID
LANSendDone( PENET_INFO pEthNetInfo )
{
    UINT32  Status;
    UINT32  Head, Tail;

    Status = pEMAC->EMAC_TSR;

    pEMAC->EMAC_TSR = ~0;

DEBUG2(LAN_DBG_LVL,"LANSendDone :Status = %x \n", Status);

    if( Status & MACB_BIT(UND) )
    {
        // reset the buffer
        INT32           i;
        PMACB_DMADESC   pRing;

        for (i = 0; i < TX_RING_SIZE; i++)
        {
            pRing = pEthNetInfo->pTxRing + i;
    
            pRing->Addr = 0;
            pRing->Ctrl = MACB_BIT(TX_USED);
        }
        pRing = pEthNetInfo->pTxRing + (TX_RING_SIZE - 1);
        pRing->Ctrl |= MACB_BIT(TX_WRAP);
    
        pEthNetInfo->TxHead = pEthNetInfo->TxTail = 0;

    }

    if( !(Status & MACB_BIT(COMP)) )
        return;

    Head = pEthNetInfo->TxHead;
    for( Tail = pEthNetInfo->TxTail; Tail != Head; Tail = NEXT_TX(Tail) )
    {
        //PENET_RINGINFO  pRingInfo = pEthNetInfo->TxRingInfo;
        PMACB_DMADESC   pTxRing;
        UINT32          BufState;

        pTxRing = pEthNetInfo->pTxRing + Tail;
        BufState = pTxRing->Ctrl;
        if( !(BufState &  MACB_BIT(TX_USED)) )
            break;

        pEthNetInfo->Stat.nTxPacket++;
        //pEthNetInfo->Stat.nTxByte += Len;
        //bp->stats.tx_packets++;
        //bp->stats.tx_bytes += skb->len;
        //rp->skb = NULL;
        //dev_kfree_skb_irq(skb);
    }
    
    pEthNetInfo->TxTail = Tail;
    //if( netif_queue_stopped(bp->dev) && (TX_BUFFS_AVAIL(bp) > MACB_TX_WAKEUP_THRESH) )
    //    netif_wake_queue(bp->dev);
DEBUG2(LAN_DBG_LVL,"LANSendDone : Tail = %d \n", Tail);

    return;
}


STATIC INT32
MACB_RxFrame( PENET_INFO pEthNetInfo, UINT32 FirstFrag, UINT32 LastFrag )
{
    UINT32          Length;
    UINT32          Frag;
    UINT32          Offset;
    PBYTE           pRxBuffer;
    PMACB_DMADESC   pRxRing;


DEBUG2(LAN_DBG_LVL,"MACB_RxFrame : FirstFrag= %d\n", FirstFrag);
DEBUG2(LAN_DBG_LVL,"MACB_RxFrame : LastFrag = %d\n", LastFrag);

    pRxRing = pEthNetInfo->pRxRing + LastFrag;
    Length = MACB_BFEXT( RX_FRMLEN, pRxRing->Ctrl );
    pRxBuffer = pEthNetInfo->RxPacket + RX_OFFSET;
    Offset = 0;
//DEBUG2(LAN_DBG_LVL,"MACB_RxFrame : pRxBuffer = %x\n", pRxBuffer);
//DEBUG2(LAN_DBG_LVL,"MACB_RxFrame : Length    = %d\n", Length);
    for( Frag = FirstFrag; ; Frag = NEXT_RX(Frag) )
    {
        UINT32  FragLen = RX_BUFFER_SIZE;

        if( Offset + FragLen > Length )
        {
            FragLen = Length - Offset;
        }

        MemCpy( pRxBuffer + Offset, pEthNetInfo->pRxBuffer + (Frag<<RX_BUFFER_SIZE_SHIFT), FragLen );
        Offset += RX_BUFFER_SIZE;

        pRxRing = pEthNetInfo->pRxRing + Frag;
        pRxRing->Addr &= ~MACB_BIT(RX_USED);

        if( Frag == LastFrag )
            break;
    }

//{int i; for(i = 0; i<Length; i++) DEBUGMSG(LAN_DBG_LVL,"%2.2x ",*(pRxBuffer + i) ); DEBUGMSG(LAN_DBG_LVL,"\n");}
    
    NwPacketInput( pEthNetInfo->iNetIf, pRxBuffer, Length );

    //RxPackets++;
    //RxBytes += Length;


    pEthNetInfo->Stat.nRxPacket++;
    pEthNetInfo->Stat.nRxByte += Length;

//DEBUG2(LAN_DBG_LVL,"MACB_RxFrame : pEthNetInfo->Stat.nRxPacket = %d\n", pEthNetInfo->Stat.nRxPacket);
    //netif_receive_skb(skb);

    return 0;
}


/* Mark DMA descriptors from begin up to and not including end as unused */
STATIC VOID
DiscardPartialFrame(PENET_INFO pEthNetInfo, UINT32 FirstFrag, UINT32 LastFrag)
{
    UINT32          iFlag;
    PMACB_DMADESC   pRxRing;


//DEBUG2(LAN_DBG_LVL,"DiscardPartialFrame: %d - %d\n",FirstFrag, LastFrag );

    for( iFlag = FirstFrag; iFlag != LastFrag; iFlag = NEXT_RX(iFlag) )
    {
        pRxRing = pEthNetInfo->pRxRing + iFlag;
        pRxRing->Addr &= ~MACB_BIT(RX_USED);
        pRxRing->Ctrl = 0;
    }

    /*
     * When this happens, the hardware stats registers for
     * whatever caused this is updated, so we don't have to record
     * anything.
     */

    return;
}

STATIC UINT32 MACB_Rx( PENET_INFO pEthNetInfo )
{
    UINT32  Addr, Ctrl;
    UINT32  ReceivedPacket = 0;
    INT32   iSOF = -1;
    UINT32  Tail = pEthNetInfo->RxTail;
    PMACB_DMADESC   pRxRing;


    for( ; ; Tail = NEXT_RX(Tail) )
    {

        pRxRing = pEthNetInfo->pRxRing + Tail;

        Addr = pRxRing->Addr;
        Ctrl = pRxRing->Ctrl;
//DEBUG2(LAN_DBG_LVL,"MACB_Rx : Addr = %x\n", Addr);
//DEBUG2(LAN_DBG_LVL,"MACB_Rx : Ctrl = %x\n", Ctrl);
        if( !(Addr & MACB_BIT(RX_USED)) )
            break;

        if( Ctrl & MACB_BIT(RX_SOF) )
        {
            if( iSOF != -1 )
                DiscardPartialFrame(pEthNetInfo, iSOF, Tail);
            iSOF = Tail;
        }

        if( Ctrl & MACB_BIT(RX_EOF) )
        {
            INT32   Dropped;

//DEBUG2(LAN_DBG_LVL,"MACB_Rx : iSOF, Tail = %d - %d\n", iSOF, Tail);
            Dropped = MACB_RxFrame(pEthNetInfo, iSOF, Tail);
            iSOF = -1;
            if( !Dropped )
            {
                ReceivedPacket++;
            }
        }
    }

    if( iSOF != -1 )
        pEthNetInfo->RxTail = iSOF;
    else
        pEthNetInfo->RxTail = Tail;


    return ReceivedPacket;
            
}



STATIC VOID
LANReceive( PENET_INFO pEthNetInfo )
{
    UINT32  Status;

    Status = pEMAC->EMAC_RSR;
    pEMAC->EMAC_RSR = Status;
DEBUG2(LAN_DBG_LVL,"LANReceive : EMAC_RSR = %x\n", Status);
    if( !Status )
    {
        /*
         * This may happen if an interrupt was pending before
         * this function was called last time, and no packets
         * have been received since.
         */
        //netif_rx_complete(dev);
        goto OUT_LAN_RECEIVE;
    }

    if( !(Status & MACB_BIT(REC)) )
    {
        DEBUG1(LAN_DBG_LVL,"No RX buffers complete\n");
        //netif_rx_complete(dev);
        goto OUT_LAN_RECEIVE;
    }

    MACB_Rx( pEthNetInfo );
    //work_done = MACB_Rx( pEthNetInfo );
    //if (work_done < orig_budget)
    //{
    //    netif_rx_complete(dev);
    //    retval = 0;
    //}
    //else
    //{
    //    retval = 1;
    //}

    /*
     * We've done what we can to clean the buffers. Make sure we
     * get notified when new packets arrive.
     */
OUT_LAN_RECEIVE:
    pEMAC->EMAC_IER = MACB_RX_INT_FLAGS;
    
    return;
}


STATIC VOID
LANISR( PENET_INFO pEthNetInfo )
{
    VOLATILE UINT32  IntStatus;
    
    IntStatus = pEMAC->EMAC_ISR;
    if( IntStatus == 0 )
        return;

    while( IntStatus )
    {
        //if( IntStatus & MACB_BIT(MFD) )
        //    Complete(&pEthNetInfo->mdio_complete);

        //if(NetIfRunning())
        //    pEMAC->EMAC_IDR = ~0;

        // Rx
        if( IntStatus & MACB_RX_INT_FLAGS )
        {
            //DEBUG2(LAN_DBG_LVL,"LANISR : LANReceive = %x\n", nRxDone );
            LANReceive( pEthNetInfo );
        }
        
        // Tx done
        if( IntStatus & (MACB_BIT(TCOMP) | MACB_BIT(ISR_TUND)) )
        {
            //DEBUG2(LAN_DBG_LVL,"LANISR : LANSendDone = %x\n", nTxDone );
            LANSendDone( pEthNetInfo );
        }

        // remained op
        if( IntStatus & MACB_BIT(HRESP) )
        {
            /*
            * TODO: Reset the hardware, and maybe move the printk
            * to a lower-priority context as well (work queue?)
            */
            DEBUG1(LAN_DBG_LVL,"DMA bus error: HRESP not OK\n");
        }

        // re-read the status register for next interrupt
        IntStatus = pEMAC->EMAC_ISR;
    }

    return;
}


//    struct net_device *dev = dev_id;
//    struct macb *bp = netdev_priv(dev);
//    u32 status;
//
//    status = macb_readl(bp, ISR);
//
//    if (unlikely(!status))
//        return IRQ_NONE;
//
//    spin_lock(&bp->lock);
//
//    while (status) {
//        if (status & MACB_BIT(MFD))
//            complete(&bp->mdio_complete);
//
//        /* close possible race with dev_close */
//        if (unlikely(!netif_running(dev))) {
//            macb_writel(bp, IDR, ~0UL);
//            break;
//        }
//
//        if (status & MACB_RX_INT_FLAGS) {
//            if (netif_rx_schedule_prep(dev)) {
//                /*
//                 * There's no point taking any more interrupts
//                 * until we have processed the buffers
//                 */
//                macb_writel(bp, IDR, MACB_RX_INT_FLAGS);
//                dev_dbg(&bp->pdev->dev, "scheduling RX softirq\n");
//                __netif_rx_schedule(dev);
//            }
//        }
//
//        if (status & (MACB_BIT(TCOMP) | MACB_BIT(ISR_TUND)))
//            macb_tx(bp);
//
//        /*
//         * Link change detection isn't possible with RMII, so we'll
//         * add that if/when we get our hands on a full-blown MII PHY.
//         */
//
//        if (status & MACB_BIT(HRESP)) {
//            /*
//             * TODO: Reset the hardware, and maybe move the printk
//             * to a lower-priority context as well (work queue?)
//             */
//            printk(KERN_ERR "%s: DMA bus error: HRESP not OK\n",
//                   dev->name);
//        }
//
//        status = macb_readl(bp, ISR);
//    }
//
//    spin_unlock(&bp->lock);
//
//    return IRQ_HANDLED;
//    
//    
//    
//    
//
//
//    UINT32          i;
//    UINT            IntStatus, IntEnabled;
//    PENET_BUFDESC   pBufferToProcess = NULL;
//    PENET_BUFDESC   pBufferProcessed = NULL;
//
//    UINT32          nRxDone, nRxNotTaken, TxFifoLevel;
//    VOLATILE UINT32 Temp;
//
//    // read device interrupt status
//    MAC_REG_READ( pEthNetInfo, INT_SRC, IntStatus);
//    //MAC_REG_WRITE( pEthNetInfo, INT_SRC, IntStatus);
//    MAC_REG_READ( pEthNetInfo, HBP_MAC_INT_ENABLE_REG, IntEnabled );
//    IntStatus &= IntEnabled;
//
////DEBUG2(LAN_DBG_LVL,"LANISR : IntEnabled = %x\n", IntEnabled );
////DEBUG2(LAN_DBG_LVL,"LANISR : IntStatus = %x\n", IntStatus );
//    // quick exit if device did not generate interrupt
//    if( IntStatus == 0)
//    {
//        goto INT_RETURN;
//        //InterruptDone( SYSINTR_LAN );
//        //return;
//    }
//
//    // handle receiver interrupt
//    if( IntStatus & (MAC_INT_RX_DONE|MAC_INT_RX_FRAME_CUT) )
//    {
////DEBUG2(LAN_DBG_LVL,"LANISR : MAC_INT_RX_DONE|MAC_INT_RX_FRAME_CUT\n" );
//        if( IntStatus & MAC_INT_RX_DONE )
//        {
//            MAC_REG_WRITE( pEthNetInfo, INT_SRC, MAC_INT_RX_DONE);
//        }
//        
//        if( IntStatus & MAC_INT_RX_FRAME_CUT )
//        {
//            MAC_REG_WRITE( pEthNetInfo, INT_SRC, MAC_INT_RX_FRAME_CUT );
//        }
//
//        // check Rx buffer level
//        MAC_REG_READ( pEthNetInfo, RX_BUFLVL, Temp );
//        nRxDone = Temp & 0x3F;
//        nRxNotTaken = (Temp & 0x3F00) >> 8;
//
////DEBUG2(LAN_DBG_LVL,"LANISR : nRxDone = %x\n", nRxDone );
//        if (nRxDone == 0)
//        {
//            goto INT_RETURN;
//            //SYS_INT_ACK (priv->ivec);
//            //spin_unlock (&priv->lock);
//            //return;
//        }
//
//        for( i = 0; i < nRxDone; i++ )
//        {
//            // obtain the current buffer address
//            pBufferProcessed = pEthNetInfo->pNextRxDone;
//            
//            MAC_REG_READ(pEthNetInfo, RX_DONE_BADDR, Temp );
//            MAC_REG_READ(pEthNetInfo, RX_BSTAT, Temp );
//            pBufferProcessed->FrameLength = ((Temp >> 16) & 0xFFFF);
//            pBufferProcessed->FrameError = (Temp & 0x5);
//
//            pBufferProcessed->Status = _B_FILLED_;
//            pEthNetInfo->pNextRxDone = pBufferProcessed->pNext;
//        }
//
//        for (i = 0; i < MAC_RX_BADDR_REG_FIFO_DEPTH - nRxNotTaken; i++)
//        {
//            pBufferToProcess = pEthNetInfo->pNextDeviceRx;
//            if( pBufferToProcess->Status == _B_FREE_ )    // there exists free buffer(s)
//            {
//                MAC_REG_WRITE( pEthNetInfo, RX_BADDR, (UINT32) pBufferToProcess->Buf);
//                pEthNetInfo->pNextDeviceRx = pBufferToProcess->pNext;
//
//                //new_skb = dev_alloc_skb (FRAME_MAX_AL);
//                //
//                //if (new_skb != NULL)
//                //{
//                //    new_skb->data += PAYLOAD_ALIGN_OFFSET;
//                //    new_skb->tail += PAYLOAD_ALIGN_OFFSET;
//                //
//                //    bufferToProcess->skb = new_skb;
//                //    bufferToProcess->pCluster = new_skb->data - PAYLOAD_ALIGN_OFFSET;
//                //
//                //    /* link new buffer to the device -> it will awake the device */
//                //    MAC_REG_WRITE (priv, RX_BADDR, (UINT) bufferToProcess->pCluster);
//                //
//                //    /* update next processing pointer (advance the pointer) */
//                //    priv->nextDeviceRx = bufferToProcess->next;
//                //}
//                //else
//                //{
//                //    break;
//                //}
//            }
//            else            /* no free buffer -> reuse the current buffer */
//            {
//                break;
//            }
//        }
//
//        /* Clear Rx Frame Drop Interrupt */
//        if( IntStatus & MAC_INT_RX_FRAME_DROPPED )
//        {
//            MAC_REG_WRITE( pEthNetInfo, INT_SRC, MAC_INT_RX_FRAME_DROPPED );
//            IntStatus &= ~(MAC_INT_RX_FRAME_DROPPED);
//        }
//
////DEBUG2(LAN_DBG_LVL,"LANISR : LANReceive = %x\n", nRxDone );
//        LANReceive( pEthNetInfo );
//        //hbp_eth_RecvHandler (dev);
//    }
//
//    // handle transmitter interrupt
//    if( IntStatus & MAC_INT_TX_DONE )
//    {
////DEBUG2(LAN_DBG_LVL,"LANISR : MAC_INT_TX_DONE\n" );
//        // clear TX_DONE INT
//        MAC_REG_WRITE( pEthNetInfo, INT_SRC, MAC_INT_TX_DONE );
//
//        if( pEthNetInfo->TxStatus == TRUE )
//        {
//            // Read number of BADDR remaining in TX_BADDR
//            MAC_REG_READ( pEthNetInfo, TX_BUFLVL, Temp);
//            TxFifoLevel = (Temp & 0x3F00) >> 8;
//
//            /* nTxFrame = nTxDone + TxFifoLevel + 1                 */
//            /* nTxDoneN is calculated approximately with safety.    */
//            /* Actual Tx Done Number may be ntxDone or nTxDone+1    */
//            pEthNetInfo->nTxDone = pEthNetInfo->nTxFrame - TxFifoLevel - 1;
//            /* Though nTxFrame can be overflowed, it is not a problem               */
//            /* becuase nTxFrame is type [unsigned int] and 0xFFFFFFFF + 1 = 0       */
//            /*  0 - 1 = 0xFFFFFFFF, 0 - 2 = 0xFFFFFFFE  , ...                       */
//
//            /* Free Tx done buffer to be used again */
//            while( pEthNetInfo->nTxFree != pEthNetInfo->nTxDone )
//            {
//                pBufferProcessed = pEthNetInfo->pNextTxDone;
//
//                MAC_REG_READ( pEthNetInfo, TX_DONE_BADDR, Temp );
//
//                //pEthNetInfo->Stats.Tx_packets++;
//                //dev_kfree_skb_irq (bufferProcessed->skb);
//                pBufferProcessed->Status = _B_FREE_;
//                pEthNetInfo->pNextTxDone = pBufferProcessed->pNext;
//                pEthNetInfo->nTxFree++;
//            }
//
//            for (i = 0; i < MAC_TX_BADDR_REG_FIFO_DEPTH - TxFifoLevel; i++)
//            {
//                pBufferToProcess = pEthNetInfo->pNextDeviceTx;
//            
//                // check next buffer is filled
//                // there exists buffer(s) to send
//                if( pBufferToProcess->Status == _B_FILLED_)
//                {
//                    // link new buffer to the device -> it will awake the device
//                    MAC_REG_WRITE( pEthNetInfo, TX_BADDR, (UINT)(pBufferToProcess->Buf) );
//                    MAC_REG_WRITE( pEthNetInfo, TX_LENGTH, pBufferToProcess->FrameLength );
//
//                    pEthNetInfo->nTxFrame++;
//            
//                    // update next processing pointer (advance the pointer)
//                    pEthNetInfo->pNextDeviceTx = pBufferToProcess->pNext;
//                }
//                else        /* next buffer is free buffer */
//                {
//                    if (i == 0 && TxFifoLevel == 0 )
//                    {
//                        pEthNetInfo->TxStatus = FALSE;    // transmitter idle
//                    }
//            
//                    break;
//                }
//            }
//        }
//        else
//        {
//            /* NONE */
//            /* Ignore Tx Done INT after changed to priv->txStatus = FALSE */
//        }
//    }
//
//    // handle dropped 
//    if( IntStatus & MAC_INT_RX_FRAME_DROPPED )
//    {
////DEBUG2(LAN_DBG_LVL,"LANISR : MAC_INT_RX_FRAME_DROPPED\n" );
//        MAC_REG_WRITE( pEthNetInfo, INT_SRC, MAC_INT_RX_FRAME_DROPPED );
//        pBufferToProcess = pEthNetInfo->pNextDeviceRx;
//
//        if( pBufferToProcess->Status == _B_FREE_ )  // there exists free buffer(s)
//        {
//            /* link new buffer to the device -> it will awake the device */
//            MAC_REG_WRITE( pEthNetInfo, RX_BADDR, (UINT32)pBufferToProcess->Buf );
//
//            /* update next processing pointer (advance the pointer) */
//            pEthNetInfo->pNextDeviceRx = pBufferToProcess->pNext;
//
//            //new_skb = dev_alloc_skb (FRAME_MAX_AL);
//            //
//              //        if (new_skb != NULL)
//            //                {
//              //                new_skb->data += PAYLOAD_ALIGN_OFFSET;
//              //                new_skb->tail += PAYLOAD_ALIGN_OFFSET;
//              //                bufferToProcess->skb = new_skb;
//              //                bufferToProcess->pCluster = new_skb->data - PAYLOAD_ALIGN_OFFSET;
//            //
//              //                /* link new buffer to the device -> it will awake the device */
//              //                MAC_REG_WRITE (priv, RX_BADDR, (UINT) bufferToProcess->pCluster);
//            //
//              //                /* update next processing pointer (advance the pointer) */
//              //                priv->nextDeviceRx = bufferToProcess->next;
//            //                }
//              //        else
//            //                {
//              //                /* Do Nothing */
//            //                }
//        }
//        else
//        {
//            // Disable MAC_INT_RX_FRAME_DROPPED
//            MAC_REG_READ( pEthNetInfo, HBP_MAC_INT_ENABLE_REG, Temp);
//            MAC_REG_WRITE( pEthNetInfo, HBP_MAC_INT_ENABLE_REG, Temp & ~(MAC_INT_RX_FRAME_DROPPED) );
//        }
//
//        //hbp_eth_RecvHandler (dev);
//        LANReceive( pEthNetInfo );
//    }
//
//      // MII
//      if( IntStatus & MAC_INT_MII_PHY )
//    {
////DEBUG2(LAN_DBG_LVL,"LANISR : MAC_INT_MII_PHY\n" );
//        if( pEthNetInfo->PhyType == PHY_KS8737 )
//        {
//            UINT32  PhyIntStatus;
//
//            // read PHY INT status register
//            // PHY interrupt will be automatically cleared by reading register
//            PhyIntStatus = PhyRegRead( pEthNetInfo, PHY_KS8737_INT_ENABLE_STATUS);
//
//            if( PhyIntStatus & (PHY_KS8737_INT_LINK_UP|PHY_KS8737_INT_LINK_DOWN) )
//            {
//                if ( /* check if eth attached ? */ 1)
//                {
//                    if(PhyLinkStatusRead( pEthNetInfo ) )
//                    {
//                        /* Link UP */
//                        //netif_carrier_on (dev);
//                    }    
//                    else
//                    {
//                        /* Link Down */
//                        //netif_carrier_off (dev);
//                    }
//                }
//            }
//        }
//    
//        // clear MII_PHY Int bit of MAC
//        MAC_REG_WRITE( pEthNetInfo, INT_SRC, MAC_INT_MII_PHY );
//    }
//
//    // fatal error
//    if( IntStatus & MAC_INT_FATAL_ERROR )
//    {
////DEBUG2(LAN_DBG_LVL,"LANISR : MAC_INT_FATAL_ERROR\n" );
//        MAC_REG_WRITE( pEthNetInfo, INT_SRC, IntStatus & MAC_INT_FATAL_ERROR );
//    }
//
//INT_RETURN:
////    MAC_REG_READ( pEthNetInfo, INT_SRC, IntStatus);
////DEBUG2(LAN_DBG_LVL,"LANISR : IntStatus = %x\n", IntStatus );
////    MAC_REG_WRITE( pEthNetInfo, INT_SRC, IntStatus);
//
//    return;
//    //SYS_INT_ACK (priv->ivec);
//    //spin_unlock (&priv->lock);
//}


STATIC BOOL
thLANDeviceDriver( UINT32 Arg )
{

    JB_RECVMESSAGE      Message;
    HANDLE              hPort;
    //HANDLE              Handle = NULL;
    PENET_INFO          pEthNetInfo;
    NET_PDEV			pNetDev = (NET_PDEV)Arg;

    hPort = PortGet( );    
    PortSetType( hPort, PORTTYPE_MESSAGE|PORTTYPE_INTERRUPT|PORTTYPE_SIGNAL );

    LAN_DeviceConfig.Name[3] = (UINT32)LAN_DeviceConfig.Minor + '1';
    LAN_DeviceConfig.Channel = ChannelGet( hPort );

    DeviceRegister( &LAN_DeviceConfig );
DEBUG2(LAN_DBG_LVL,"thLANDeviceDriver pDeviceConfig->Channel = %X\n",LAN_DeviceConfig.Channel);

    pEthNetInfo = &EthNetInfo;
    //pEthNetInfo->iDev = iDev;

DEBUG2(LAN_DBG_LVL,"thLANDeviceDriver LANInit pNetDev = %x\n",pNetDev );
DEBUG2(LAN_DBG_LVL,"thLANDeviceDriver LANInit &LAN_NetDev = %x\n",&LAN_NetDev );


DEBUG2(LAN_DBG_LVL,"thLANDeviceDriver LANInit pEthNetInfo = %X\n",pEthNetInfo );
    LANInit( pEthNetInfo, pNetDev);

DEBUG2(LAN_DBG_LVL,"thLANDeviceDriver InterruptRegister pEthNetInfo->nINTVirt = %d\n",pEthNetInfo->nINTVirt );
    InterruptRegister( pEthNetInfo->nINTVirt, hPort);

    // open routine
DEBUG2(LAN_DBG_LVL,"thLANDeviceDriver LANOpen pEthNetInfo = %X\n",pEthNetInfo );
    LANOpen( pEthNetInfo );

    pEthNetInfo->Handle = ChannelOpenHandleGetFrom( LAN_DeviceConfig.Channel );
    DEBUG2(LAN_DBG_LVL,"thDeviceDriverThreadHead Handle = %X\n",pEthNetInfo->Handle);

    pNetDev->pDrvName = &LAN_DeviceConfig.Name[0];
    pNetDev->Handle = pEthNetInfo->Handle;

    pEthNetInfo->iNetIf = NwRegisterDev( pNetDev );

    InterruptEnable(pEthNetInfo->nINTVirt);

    while(1)
    {
        MessageReceive( hPort, &Message);
DEBUGMSG(LAN_DBG_LVL,"thLANDeviceDriver : MessageReceive\n" );

        switch(Message.Type)
        {
        case MESSAGE_TYPE_SIGNAL:
        {
            if( Message.Param & SIG_RESERVED_ALARM)
            {
            }

            if( Message.Param & SIG_NETWORK_OUTPUT)
            {
                BOOL            bRet;

DEBUGMSG(LAN_DBG_LVL,"thLANDeviceDriver : MESSAGE_TYPE_SIGNAL %d\n", pEthNetInfo->iDev);
NEXT_PACKET:
                bRet = LANSend( pEthNetInfo );
DEBUGMSG(LAN_DBG_LVL,"thLANDeviceDriver : MESSAGE_TYPE_SIGNAL bRet = %d\n", bRet);
                if( bRet )
                    goto NEXT_PACKET;
            }

            if( Message.Param & SIG_NETWORK_SET)
            {
                NET_DEV_SET   NetDevSet;
                
                UINT32  Resp = -ERR_NETWORK_OPNOTSUPP;
DEBUGMSG( LAN_DBG_LVL,"%s (%d): SIG_NETWORK_SET Resp  %x\n", __FUNCTION__, __LINE__, Resp);
                // get set data from network server
                if( NwDevGetSet( pNetDev->iNetIf, &NetDevSet, sizeof(NET_DEV_SET) ) )
                {
                    // response the set data 
                    NwDevRespSet( pEthNetInfo->iNetIf, Resp );
                }
            }

#if 0 // case of wireless network 
            if( Message.Param & SIG_NETWORK_SET)
            {
                NET_DEV_SET   NetDevSet;
                
                UINT32  Resp = -EOPNOTSUPP;
                // get the network set
                struct ifreq *p_ifreq;
                UINT32          Cmd;

                // get set data from network server
                if( NwDevGetSet( pNetDev->iNetIf, &NetDevSet, sizeof(NET_DEV_SET) ) )
                {
DEBUGMSG( DBG_LVL_WLAN_SET,"%s (%d): SIG_NETWORK_SET Cmd = %x\n", __FUNCTION__, __LINE__, NetDevSet.Cmd);
//kprintf("%s (%d): SIG_NETWORK_SET Cmd = %x\n", __FUNCTION__, __LINE__, NetDevSet.Cmd);
                    /* check the wireless standard control */
                    if( NetDevSet.Cmd >= SOCK_CTRL_WLAN_FIRST && 
                        NetDevSet.Cmd <= SOCK_CTRL_WLAN_LAST)
                    {

#if 1
                        Resp = 
                        wireless_process_ioctl(net_dev, (struct ifreq *)NetDevSet.pSet, NetDevSet.Cmd);
#else
                        iw_handler  handler;
                        struct iw_handler_def *pwh;
                        pwh = net_dev->wireless_handlers;
                        
                        if( pwh->standard == NULL ||
                            IW_IOCTL_IDX(NetDevSet.Cmd) >= pwh->num_standard)
                            goto SET_DONE;

                        PNWDEVREQ pNwDevReq = (PNWDEVREQ)NetDevSet.pSet;
                    	struct iw_request_info info = { .cmd = NetDevSet.Cmd, .flags = 0 };

                        handler = pwh->standard[IW_IOCTL_IDX(NetDevSet.Cmd)];

                        Resp = handler(net_dev, &info, &pNwDevReq->NwDevReqU, NULL/*extra*/);
#endif
                    }
                    else
                    if( NetDevSet.Cmd >= SOCK_CTRL_WLAN_FIRST && 
                        NetDevSet.Cmd <= SOCK_CTRL_WLAN_LAST)
                    {
#if 1
                        Resp = 
                        wireless_process_ioctl(net_dev, (struct ifreq *)NetDevSet.pSet, NetDevSet.Cmd);
#else
                        iw_handler  handler;
                        struct iw_handler_def *pwh;
                        pwh = net_dev->wireless_handlers;
                        
                        if( pwh->private == NULL ||
                            IW_IOCTL_IDX(NetDevSet.Cmd) >= pwh->num_private)
                            goto SET_DONE;

                        PNWDEVREQ pNwDevReq = (PNWDEVREQ)NetDevSet.pSet;
                    	struct iw_request_info info = { .cmd = NetDevSet.Cmd, .flags = 0 };

                        handler = pwh->standard[IW_IOCTL_IDX(NetDevSet.Cmd)];

                        Resp = handler(net_dev, &info, &pNwDevReq->NwDevReqU, NULL/*extra*/);
#endif
                    }
                    else
                    {
                        Resp = 
                        net_dev->do_ioctl( net_dev, (struct ifreq *)NetDevSet.pSet, NetDevSet.Cmd );
                    }
SET_DONE:
DEBUGMSG( DBG_LVL_WLAN_SET,"%s (%d): SIG_NETWORK_SET Resp  %x\n", __FUNCTION__, __LINE__, Resp);
//kprintf("%s (%d): SIG_NETWORK_SET Resp  %x\n", __FUNCTION__, __LINE__, Resp);
                    // response the set data 
                    NwDevRespSet( pNetDev->iNetIf, Resp );
DEBUGMSG( DBG_LVL_WLAN_SET,"%s (%d): SIG_NETWORK_SET ++\n", __FUNCTION__, __LINE__);
                }
DEBUGMSG( DBG_LVL_WLAN_SET,"%s (%d): SIG_NETWORK_SET ++\n", __FUNCTION__, __LINE__);
            }
#endif
            continue;
        }

        case MESSAGE_TYPE_INTERRUPT:
        {
DEBUG2(LAN_DBG_LVL,"thLANDeviceDriver : MESSAGE_TYPE_INTERRUPT %d\n", pEthNetInfo->iDev);
            LANISR( pEthNetInfo );
            InterruptDone( pEthNetInfo->nINTVirt );
            continue;
        }

        case MESSAGE_TYPE_DEVICE_READ:
        {
            Message.Type = 0;
            break;
        }

        case MESSAGE_TYPE_DEVICE_WRITE:
        {
            Message.Type = 0;
            break;
        }

        case MESSAGE_TYPE_DEVICE_CONTROL:
        {
            Message.Type = FALSE;
            break;
        }

        case MESSAGE_TYPE_DEVICE_OPEN:
        {
            //DEBUG2(1,"MESSAGE_TYPE_DEVICE_OPEN %X\n", Channel);
            //if( Handle != NULL)
            //{
            //    ThreadSetLastError( Message.hSender, ERR_BUSY );
            //    Message.Type = NULL;
            //    break;
            //}
            //
            //LANOpen();
            //
            //Handle = ChannelOpenHandleGetFrom( Channel );
            //DEBUG2(1,"thDeviceDriverThreadHead %X\n",Handle);
            //
            ////InterruptEnable(SYSINTR_LAN);
            //
            //Message.Type = Handle;

            Message.Type = NULL;
            break;
        }

        case MESSAGE_TYPE_DEVICE_CLOSE:
        {
            //DEBUG2(1,"MESSAGE_TYPE_DEVICE_CLOSE %X\n", Handle);
            //DEBUG2(1,"MESSAGE_TYPE_DEVICE_CLOSE %X\n", Message.Handle);

            if( Message.Handle != pEthNetInfo->Handle )
            {
                ThreadSetLastError( Message.hSender, ERR_INVALID );
                Message.Type = FALSE;
                break;
            }

            LANClose( pEthNetInfo );
            InterruptDisable( pEthNetInfo->nINTVirt );

            pEthNetInfo->Handle = NULL;
            Message.Type = TRUE;
            break;
        }

        default:
            break;
        }

        MessageReply(&Message);

    }

    return TRUE;    
}




#include <moduleloader.h>

PUBLIC HANDLE
DllMain( HANDLE hDll, UINT32 Reason, PVOID Reserved )
{
    HANDLE hThread = NULL;
    switch( Reason )
    {
    case DLL_PROCESS_ATTACH:
        hThread = ThreadCreate (&MACB_LANDev_Thread);
    	break;
    
    case DLL_PROCESS_DETACH:
    	break;
    }

    return hThread;
}




