#ifndef __PHY_H__
#define __PHY_H__

#include <types.h>

#define SPEED_100       100
#define SPEED_10        10

// PHY access information
typedef struct __MAC_PHY_OP__
{
    VOID (*ReadPhy)     (PVOID pDeviceLocation, BYTE PhyAddr, BYTE nReg, PUINT32 pValue);    // function to read a PHY register
    VOID (*WritePhy)    (PVOID pDeviceLocation, BYTE PhyAddr, BYTE nReg, UINT32 Value);      // function to write a PHY register
    PVOID               pDeviceLocation;    // MAC's base address
}MAC_PHY_OP, *MAC_PHY_POP;

// PHY configuration structure
typedef struct __PHY_INFO__
{
    UINT32      Speed;              // Link Speed
    BOOL        bFullDuplex;        // Full or Half Duplex
    BOOL        bAutoNegociation;   // indicates if Autonegociation is active or not
    BOOL        bRMII;              // indicates wether the PHY is used in RMII mode
} PHY_INFO, *PHY_PINFO;

typedef struct __PHY_INTERFACE__
{
    PBYTE       pName;
    BOOL (*PHY_Reset)                           (MAC_PHY_OP *pMacPhyOp,BYTE PhyAddr);
    BOOL (*PHY_SetConfiguration)                (MAC_PHY_OP *pMacPhyOp,BYTE PhyAddr, PHY_PINFO pPhyInfo);
    BOOL (*PHY_GetConfiguration)                (MAC_PHY_OP *pMacPhyOp,BYTE PhyAddr, PHY_PINFO pPhyInfo);
    BOOL (*PHY_StartAutoNegociation)            (MAC_PHY_OP *pMacPhyOp,BYTE PhyAddr);
    BOOL (*PHY_WaitForLink)                     (MAC_PHY_OP *pMacPhyOp,BYTE PhyAddr, UINT32 dwTimeout);
    BOOL (*PHY_WaitForAutonegociationComplete)  (MAC_PHY_OP *pMacPhyOp,BYTE PhyAddr, UINT32 dwTimeout);
    BOOL (*PHY_GetID)                           (MAC_PHY_OP *pMacPhyOp,BYTE PhyAddr, UINT32* pdwID);
    BOOL (*PHY_CheckID)                         (MAC_PHY_OP *pMacPhyOp,BYTE PhyAddr);
}PHY_INTERFACE, *PHY_PINTERFACE;




STATIC INLINE PHY_PINTERFACE
FindPhy( MAC_PHY_POP pMacPhyOp, PHY_PINTERFACE PhyList[], UINT32 nEntry, PBYTE pPhyAddr )
{
#define MAX_PHYADDRESS          32
    UINT32          i;
    PHY_PINTERFACE  pCurrentPhy;
    
    if (pPhyAddr == NULL)
    {
        return NULL;
    }

    for( i = 0 ; i < nEntry ; i++ )
    {
        pCurrentPhy = PhyList[i];
        for( *pPhyAddr=0; *pPhyAddr < MAX_PHYADDRESS; (*pPhyAddr)++ )
        {
            if( pCurrentPhy->PHY_CheckID( pMacPhyOp,*pPhyAddr) )
            {
                return pCurrentPhy;
            }
        }        
    }
    
    return NULL;
}

#endif //__PHY_H__

